"""Module to manange AiiDA lab configuration."""
from os import getenv

AIIDALAB_HOME = getenv('AIIDALAB_HOME', '/project')
AIIDALAB_APPS = getenv('AIIDALAB_APPS', '/project/apps')
AIIDALAB_SCRIPTS = getenv('AIIDALAB_SCRIPTS', '/opt')
AIIDALAB_REGISTRY = getenv('AIIDALAB_REGISTRY', 'https://aiidalab.materialscloud.org/appsdata/apps_meta.json')
AIIDALAB_DEFAULT_GIT_BRANCH = getenv('AIIDALAB_DEFAULT_GIT_BRANCH', 'master')



"""Helpful utilities for the AiiDA lab tools."""

import sys
import json
import time
from os import path
from importlib import import_module
from urllib.parse import urlparse
from collections import defaultdict
from functools import wraps
from threading import Lock

import requests
from markdown import markdown
import ipywidgets as ipw
from IPython.lib import backgroundjobs as bg

from .config import AIIDALAB_APPS, AIIDALAB_REGISTRY


def update_cache():
    """Run this process asynchronously."""
    requests_cache.install_cache(cache_name='apps_meta', backend='sqlite', expire_after=3600, old_data_on_error=True)
    requests.get(AIIDALAB_REGISTRY)
    requests_cache.install_cache(cache_name='apps_meta', backend='sqlite')


# Warning: try-except is a fix for Quantum Mobile release v19.03.0 that does not have requests_cache installed
try:
    import requests_cache
    # At start getting data from cache
    requests_cache.install_cache(cache_name='apps_meta', backend='sqlite')

    # If requests_cache is installed, upgrade the cache in the background.
    UPDATE_CACHE_BACKGROUND = bg.BackgroundJobFunc(update_cache)
    UPDATE_CACHE_BACKGROUND.start()
except ImportError:
    pass


def load_app_registry():
    """Load apps' information from the AiiDA lab registry."""
    parsed_url = urlparse(AIIDALAB_REGISTRY)
    if parsed_url.scheme == 'file':
        with open(parsed_url.path) as file:
            return json.loads(file.read())
    else:
        try:
            return requests.get(AIIDALAB_REGISTRY).json()
        except ValueError:
            print("Registry server is unavailable! Can't check for the updates")
            return dict(apps=dict(), catgories=dict())


def load_widget(name):
    if path.exists(path.join(AIIDALAB_APPS, name, 'start.py')):
        return load_start_py(name)
    return load_start_md(name)


def load_start_py(name):
    """Load app appearance from a Python file."""
    try:
        mod = import_module('apps.%s.start' % name)
        appbase = "../" + name
        jupbase = "../../.."
        notebase = jupbase + "/notebooks/apps/" + name
        try:
            return mod.get_start_widget(appbase=appbase, jupbase=jupbase, notebase=notebase)
        except TypeError:
            return mod.get_start_widget(appbase=appbase, jupbase=jupbase)
    except Exception:  # pylint: disable=broad-except
        return ipw.HTML("<pre>{}</pre>".format(sys.exc_info()))


def load_start_md(name):
    """Load app appearance from a Markdown file."""
    fname = path.join(AIIDALAB_APPS, name, 'start.md')
    try:

        md_src = open(fname).read()
        md_src = md_src.replace("](./", "](../{}/".format(name))
        html = markdown(md_src)

        # open links in new window/tab
        html = html.replace('<a ', '<a target="_blank" ')

        # downsize headings
        html = html.replace("<h3", "<h4")
        return ipw.HTML(html)

    except Exception as exc:  # pylint: disable=broad-except
        return ipw.HTML("Could not load start.md: {}".format(str(exc)))


class throttled:  # pylint: disable=invalid-name
    """Decorator to throttle calls to a function to a specified rate.

    The throttle is specific to the first argument of the wrapped
    function. That means for class methods it is specific to each
    instance.

    Adapted from: https://gist.github.com/gregburek/1441055

    """

    def __init__(self, calls_per_second=1):
        self.calls_per_second = calls_per_second
        self.last_start = defaultdict(lambda: -1)
        self.locks = defaultdict(Lock)

    def __call__(self, func):
        """Return decorator function."""

        @wraps(func)
        def wrapped(instance, *args, **kwargs):
            if self.last_start[hash(instance)] >= 0:
                elapsed = time.perf_counter() - self.last_start[hash(instance)]
                to_wait = 1.0 / self.calls_per_second - elapsed
                if to_wait > 0:
                    locked = self.locks[hash(instance)].acquire(blocking=False)
                    if locked:
                        try:
                            time.sleep(to_wait)
                        finally:
                            self.locks[hash(instance)].release()
                    else:
                        return None  # drop

            self.last_start[hash(instance)] = time.perf_counter()
            return func(instance, *args, **kwargs)

        return wrapped



"""Module to generate AiiDA lab home page."""
from os import path
from glob import glob
from functools import wraps

import json
import traitlets
import ipywidgets as ipw
from IPython.display import display

# AiiDA lab imports.
from .app import AiidaLabApp
from .config import AIIDALAB_APPS
from .utils import load_widget, load_app_registry
from .widgets import UpdateAvailableInfoWidget


def create_app_widget_move_buttons(name):
    """Make buttons to move the app widget up or down."""
    layout = ipw.Layout(width="40px")
    app_widget_move_buttons = ipw.HTML("""
    <a href=./start.ipynb?move_up={name} title="Move it up"><i class='fa fa-arrow-up' style='color:#337ab7;font-size:2em;' ></i></a>
    <a href=./start.ipynb?move_down={name} title="Move it down"><i class='fa fa-arrow-down' style='color:#337ab7;font-size:2em;' ></i></a>
    """.format(name=name),
                                       layout=layout)
    app_widget_move_buttons.layout.margin = "50px 0px 0px 0px"

    return app_widget_move_buttons


def _workaround_property_lock_issue(func):
    """Work-around for issue with the ipw.Accordion widget.

    The widget does not report changes to the .selected_index trait when displayed
    within a custom ipw.Output instance. However, the change is somewhat cryptic reported
    by a change to the private '_property_lock' trait. We observe changes to that trait
    and convert the change argument into a form that is more like the one expected by
    downstream handlers.
    """

    @wraps(func)
    def _inner(self, change):
        if change['name'] == '_property_lock':
            if 'selected_index' in change['old']:
                fixed_change = change.copy()
                fixed_change['name'] = 'selected_index'
                fixed_change['new'] = change['old']['selected_index']
                del fixed_change['old']
                return func(self, fixed_change)

        return func(self, change)

    return _inner


class AiidaLabHome:
    """Class that mananges the appearance of the AiiDA lab home page."""

    def __init__(self):
        self.config_fn = ".launcher.json"
        self.output = ipw.Output()
        self.app_registry = load_app_registry()['apps']
        self._app_widgets = dict()

    def _create_app_widget(self, name):
        """Create the widget representing the app on the home screen."""
        config = self.read_config()
        app_data = self.app_registry.get(name, None)
        app = AiidaLabApp(name, app_data, AIIDALAB_APPS)

        if name == 'home':
            app_widget = AppWidget(app, allow_move=False)
        else:
            app_widget = CollapsableAppWidget(app, allow_move=True)
            app_widget.hidden = name in config['hidden']
            app_widget.observe(self._on_app_widget_change_hidden, names=['hidden'])

        return app_widget

    def _on_app_widget_change_hidden(self, change):
        """Record whether a app widget is hidden on the home screen in the config file."""
        config = self.read_config()
        hidden = set(config['hidden'])
        if change['new']:  # hidden
            hidden.add(change['owner'].app.name)
        else:  # visible
            hidden.discard(change['owner'].app.name)
        config['hidden'] = list(hidden)
        self.write_config(config)

    def write_config(self, config):
        json.dump(config, open(self.config_fn, 'w'), indent=2)

    def read_config(self):
        if path.exists(self.config_fn):
            return json.load(open(self.config_fn, 'r'))
        return {'order': [], 'hidden': []}  #default config

    def render(self):
        """Rendering all apps."""
        self.output.clear_output()
        apps = self.load_apps()

        with self.output:
            for name in apps:

                # Create app widget if it has not been created yet.
                if name not in self._app_widgets:
                    self._app_widgets[name] = self._create_app_widget(name)

                display(self._app_widgets[name])

        return self.output

    def load_apps(self):
        """Load apps according to the order defined in the config file."""
        apps = [
            path.basename(fn)
            for fn in glob(path.join(AIIDALAB_APPS, '*'))
            if path.isdir(fn) and not fn.endswith('home') and not fn.endswith('__pycache__')
        ]
        config = self.read_config()
        order = config['order']
        apps.sort(key=lambda x: order.index(x) if x in order else -1)
        config['order'] = apps
        self.write_config(config)
        return ['home'] + apps

    def move_updown(self, name, delta):
        """Move the app up/down on the start page."""
        config = self.read_config()
        order = config['order']
        i = order.index(name)
        del order[i]
        j = min(len(order), max(0, i + delta))
        order.insert(j, name)
        config['order'] = order
        self.write_config(config)


class AppWidget(ipw.VBox):
    """Widget that represents an app as part of the home page."""

    def __init__(self, app, allow_move=False):
        self.app = app

        launcher = load_widget(app.name)
        launcher.layout = ipw.Layout(width="900px")

        update_info = UpdateAvailableInfoWidget()
        ipw.dlink((app, 'updates_available'), (update_info, 'updates_available'))
        update_info.layout.margin = "0px 0px 0px 800px"

        if allow_move:
            app_widget_move_buttons = create_app_widget_move_buttons(app.name)
            body = ipw.HBox([launcher, app_widget_move_buttons])
        else:
            body = launcher

        footer = ipw.HTML("<a href=./single_app.ipynb?app={}><button>Manage App</button></a>".format(app.name),
                          layout={'width': 'initial'})
        if app.url:
            footer.value += ' <a href="{}"><button>URL</button></a>'.format(app.url)
        footer.layout.margin = "0px 0px 0px 700px"

        super().__init__(children=[update_info, body, footer])


class CollapsableAppWidget(ipw.Accordion):
    """Widget that represents a collapsable app as part of the home page."""

    hidden = traitlets.Bool()

    def __init__(self, app, **kwargs):
        self.app = app
        app_widget = AppWidget(app, **kwargs)
        super().__init__(children=[app_widget])
        self.set_title(0, app.title)
        # Need to observe all names here due to unidentified issue:
        self.observe(self._observe_accordion_selected_index)  # , names=['selected_index'])

    @_workaround_property_lock_issue
    def _observe_accordion_selected_index(self, change):
        if change['name'] == 'selected_index':  # Observing all names due to unidentified issue.
            self.hidden = change['new'] is None

    @traitlets.observe('hidden')
    def _observe_hidden(self, change):
        self.selected_index = None if change['new'] else 0



# -*- coding: utf-8 -*-
"""Module to manage AiiDA lab apps."""

import re
import os
import shutil
import json
from contextlib import contextmanager
from time import sleep
from threading import Thread
from subprocess import check_output, STDOUT, CalledProcessError

import requests
import traitlets
import ipywidgets as ipw
from dulwich.porcelain import fetch
from dulwich.errors import NotGitRepository
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from jinja2 import Template

from .config import AIIDALAB_DEFAULT_GIT_BRANCH
from .widgets import StatusHTML, Spinner
from .git_util import GitManagedAppRepo as Repo
from .utils import throttled

HTML_MSG_SUCCESS = """<i class="fa fa-check" style="color:#337ab7;font-size:1em;" ></i>
{}"""

HTML_MSG_FAILURE = """<i class="fa fa-times" style="color:red;font-size:1em;" ></i>
{}"""


class AppNotInstalledException(Exception):
    pass


class VersionSelectorWidget(ipw.VBox):
    """Class to choose app's version."""

    disabled = traitlets.Bool()
    available_versions = traitlets.Dict(traitlets.Unicode())

    def __init__(self, *args, **kwargs):
        style = {'description_width': '100px'}
        self.release_line = ipw.Dropdown(
            description='Release line',
            style=style,
        )
        self.installed_version = ipw.Text(
            description='Installed version',
            disabled=True,
            style=style,
        )
        self.info = StatusHTML('')

        super().__init__(
            children=[self.release_line, self.installed_version, self.info],
            layout={'min_width': '300px'},
            *args,
            **kwargs,
        )

    @traitlets.observe('disabled')
    def _observe_disabled(self, change):
        self.release_line.disabled = change['new']


class AiidaLabApp(traitlets.HasTraits):
    """Manage installation status of an AiiDA lab app.

    Arguments:

        name (str):
            Name of the Aiida lab app.
        app_data (dict):
            Dictionary containing the app metadata.
        aiidalab_apps_path (str):
            Path to directory at which the app is expected to be installed.
    """

    path = traitlets.Unicode(allow_none=True, readonly=True)
    install_info = traitlets.Unicode()

    available_release_lines = traitlets.Set(traitlets.Unicode)
    installed_release_line = traitlets.Unicode(allow_none=True)
    installed_version = traitlets.Unicode(allow_none=True)
    updates_available = traitlets.Bool(readonly=True, allow_none=True)

    busy = traitlets.Bool(readonly=True)
    modified = traitlets.Bool(readonly=True, allow_none=True)

    class AppPathFileSystemEventHandler(FileSystemEventHandler):
        """Internal event handeler for app path file system events."""

        def __init__(self, app):
            self.app = app

        def on_any_event(self, event):
            """Refresh app for any event."""
            self.app.refresh_async()

    def __init__(self, name, app_data, aiidalab_apps_path):
        super().__init__()

        if app_data is not None:
            self._git_url = app_data['git_url']
            self._meta_url = app_data['meta_url']
            self._git_remote_refs = app_data['gitinfo']
            self.categories = app_data['categories']
            self._meta_info = app_data['metainfo']
        else:
            self._git_url = None
            self._meta_url = None
            self._git_remote_refs = {}
            self.categories = None
            self._meta_info = None

        self._observer = None
        self._check_install_status_changed_thread = None

        self.name = name
        self.path = os.path.join(aiidalab_apps_path, self.name)
        self.refresh_async()
        self._watch_repository()

    @traitlets.default('modified')
    def _default_modified(self):
        if self.is_installed():
            return self._repo.dirty()
        return None

    @traitlets.default('busy')
    def _default_busy(self):  # pylint: disable=no-self-use
        return False

    @contextmanager
    def _show_busy(self):
        """Apply this decorator to indicate that the app is busy during execution."""
        self.set_trait('busy', True)
        try:
            yield
        finally:
            self.set_trait('busy', False)

    def _watch_repository(self):
        """Watch the app repository for file system events.

        The app state is refreshed automatically for all events.
        """
        if self._observer is None and os.path.isdir(self.path):
            event_handler = self.AppPathFileSystemEventHandler(self)

            self._observer = Observer()
            self._observer.schedule(event_handler, self.path, recursive=True)
            self._observer.start()

        if self._check_install_status_changed_thread is None:

            def check_install_status_changed():
                installed = self.is_installed()
                while not self._check_install_status_changed_thread.stop_flag:
                    if installed != self.is_installed():
                        installed = self.is_installed()
                        self.refresh()
                    sleep(1)

            self._check_install_status_changed_thread = Thread(target=check_install_status_changed)
            self._check_install_status_changed_thread.stop_flag = False
            self._check_install_status_changed_thread.start()

    def _stop_watch_repository(self, timeout=None):
        """Stop watching the app repository for file system events."""
        if self._observer is not None:
            self._observer.stop()
            self._observer.join(timeout=timeout)
            if not self._observer.isAlive():
                self._observer = None

        if self._check_install_status_changed_thread is not None:
            self._check_install_status_changed_thread.stop_flag = True
            self._check_install_status_changed_thread.join(timeout=timeout)
            if not self._check_install_status_changed_thread.is_alive():
                self._check_install_status_changed_thread = None

    def __del__(self):  # pylint: disable=missing-docstring
        self._stop_watch_repository(1)

    def in_category(self, category):
        # One should test what happens if the category won't be defined.
        return category in self.categories

    def is_installed(self):
        """The app is installed if the corresponding folder is present."""
        return os.path.isdir(self.path)

    def _has_git_repo(self):
        """Check if the app has a .git folder in it."""
        try:
            Repo(self.path)
            return True
        except NotGitRepository:
            return False

    def install_app(self, version=None):
        """Installing the app."""
        assert self._git_url is not None
        if version is None:
            version = 'git:refs/heads/' + AIIDALAB_DEFAULT_GIT_BRANCH

        with self._show_busy():
            assert version.startswith('git:refs/heads/')
            branch = re.sub(r'git:refs\/heads\/', '', version)

            if not os.path.isdir(self.path):  # clone first
                check_output(['git', 'clone', '--branch', branch, self._git_url, self.path],
                             cwd=os.path.dirname(self.path),
                             stderr=STDOUT)

            check_output(['git', 'checkout', '-f', branch], cwd=self.path, stderr=STDOUT)
            self.refresh()
            self._watch_repository()
            return branch

    def update_app(self, _=None):
        """Perform app update."""
        assert self._git_url is not None
        with self._show_busy():
            fetch(repo=self._repo, remote_location=self._git_url)
            tracked_branch = self._repo.get_tracked_branch()
            check_output(['git', 'reset', '--hard', tracked_branch], cwd=self.path, stderr=STDOUT)
            self.refresh_async()

    def uninstall_app(self, _=None):
        """Perfrom app uninstall."""
        # Perform uninstall process.
        with self._show_busy():
            self._stop_watch_repository()
            try:
                shutil.rmtree(self.path)
            except FileNotFoundError:
                raise RuntimeError("App was already uninstalled!")
            self.refresh()
            self._watch_repository()

    def check_for_updates(self):
        """Check whether there is an update available for the installed release line."""
        try:
            assert self._git_url is not None
            branch_ref = 'refs/heads/' + self._repo.branch().decode()
            assert self._repo.get_tracked_branch() is not None
            remote_update_available = self._git_remote_refs.get(branch_ref) != self._repo.head().decode()
            self.set_trait('updates_available', remote_update_available or self._repo.update_available())
        except (AssertionError, RuntimeError):
            self.set_trait('updates_available', None)

    def _available_release_lines(self):
        """"Return all available release lines (local and remote)."""
        for branch in self._repo.list_branches():
            yield 'git:refs/heads/' + branch.decode()
        for ref in self._git_remote_refs:
            if ref.startswith('refs/heads/'):
                yield 'git:' + ref

    @throttled(calls_per_second=1)
    def refresh(self):
        """Refresh app state."""
        with self._show_busy():
            with self.hold_trait_notifications():
                if self.is_installed() and self._has_git_repo():
                    self.available_release_lines = set(self._available_release_lines())
                    try:
                        self.installed_release_line = 'git:refs/heads/' + self._repo.branch().decode()
                    except RuntimeError:
                        self.installed_release_line = None
                    self.installed_version = self._repo.head()
                    self.check_for_updates()
                    self.set_trait('modified', self._repo.dirty())
                else:
                    self.available_release_lines = set()
                    self.installed_release_line = None
                    self.installed_version = None
                    self.set_trait('updates_available', None)
                    self.set_trait('modified', None)

    def refresh_async(self):
        """Asynchronized (non-blocking) refresh of the app state."""
        refresh_thread = Thread(target=self.refresh)
        refresh_thread.start()

    @property
    def metadata(self):
        """Return metadata dictionary. Give the priority to the local copy (better for the developers)."""
        if self.is_installed():
            try:
                with open(os.path.join(self.path, 'metadata.json')) as json_file:
                    return json.load(json_file)
            except IOError:
                return dict()
        elif self._meta_info is not None:
            return dict(self._meta_info)
        elif self._meta_url is None:
            raise RuntimeError(
                f"Requested app '{self.name}' is not installed and is also not registered on the app registry.")
        else:
            return requests.get(self._meta_url).json()

    def _get_from_metadata(self, what):
        """Get information from metadata."""

        try:
            return "{}".format(self.metadata[what])
        except KeyError:
            if not os.path.isfile(os.path.join(self.path, 'metadata.json')):
                return '({}) metadata.json file is not present'.format(what)
            return 'the field "{}" is not present in metadata.json file'.format(what)

    @property
    def authors(self):
        return self._get_from_metadata('authors')

    @property
    def description(self):
        return self._get_from_metadata('description')

    @property
    def title(self):
        return self._get_from_metadata('title')

    @property
    def url(self):
        """Provide explicit link to Git repository."""
        return self._git_url

    @property
    def more(self):
        return """<a href=./single_app.ipynb?app={}>Manage App</a>""".format(self.name)

    @property
    def logo(self):
        """Return logo object. Give the priority to the local version"""

        # For some reason standard ipw.Image() app does not work properly.
        res = ipw.HTML('<img src="./aiidalab_logo_v4.svg">', layout={'width': '100px', 'height': '100px'})

        # Checking whether the 'logo' key is present in metadata dictionary.
        if 'logo' not in self.metadata:
            res.value = '<img src="./aiidalab_logo_v4.svg">'

        # If 'logo' key is present and the app is installed.
        elif self.is_installed():
            res.value = '<img src="{}">'.format(os.path.join('..', self.name, self.metadata['logo']))

        # If not installed, getting file from the remote git repository.
        else:
            # Remove .git if present.
            html_link = os.path.splitext(self._git_url)[0]

            # We expect it to always be a git repository
            html_link += '/master/' + self.metadata['logo']
            if 'github.com' in html_link:
                html_link = html_link.replace('github.com', 'raw.githubusercontent.com')
                if html_link.endswith('.svg'):
                    html_link += '?sanitize=true'
            res.value = '<img src="{}">'.format(html_link)

        return res

    @property
    def _repo(self):
        """Returns Git repository."""
        if not self.is_installed():
            raise AppNotInstalledException("The app is not installed")
        return Repo(self.path)

    def render_app_manager_widget(self):
        """Display widget to manage the app."""
        try:
            return AppManagerWidget(self, with_version_selector=True)
        except Exception as error:  # pylint: disable=broad-except
            return ipw.HTML(
                '<div style="font-size: 30px; text-align:center;">'
                f'Unable to show app widget due to error: {error}'
                '</div>',
                layout={'width': '600px'})


class AppManagerWidget(ipw.VBox):
    """Widget for management of apps.

    Shows basic information about the app (description, authors, etc.) and provides
    an interface to install, uninstall, and update the application, as well as change
    versions if possible.
    """

    TEMPLATE = Template("""<b> <div style="font-size: 30px; text-align:center;">{{ app.title }}</div></b>
    <br>
    <b>Authors:</b> {{ app.authors }}
    <br>
    <b>Description:</b> {{ app.description }}
    {% if app.url %}
    <br>
    <b>URL:</b> <a href="{{ app.url }}">{{ app.url }}</a>
    {% endif %}""")

    def __init__(self, app, with_version_selector=False):
        self.app = app

        body = ipw.HTML(self.TEMPLATE.render(app=app))
        body.layout = {'width': '600px'}

        # Setup install_info
        self.install_info = StatusHTML()

        # Setup buttons
        self.install_button = ipw.Button(description='Install', disabled=True)
        self.install_button.on_click(self._install_version)

        self.uninstall_button = ipw.Button(description='Uninstall', disabled=True)
        self.uninstall_button.on_click(self._uninstall_app)

        self.update_button = ipw.Button(description='Update', disabled=True)
        self.update_button.on_click(self._update_app)

        self.modifications_indicator = ipw.HTML()
        self.modifications_ignore = ipw.Checkbox(description="Ignore")
        self.modifications_ignore.observe(self._refresh_widget_state)

        self.spinner = Spinner("color:#337ab7;font-size:1em;")
        ipw.dlink((self.app, 'busy'), (self.spinner, 'enabled'))

        children = [
            ipw.HBox([app.logo, body]),
            ipw.HBox([self.uninstall_button, self.install_button, self.update_button, self.spinner]),
            ipw.HBox([self.install_info]),
            ipw.HBox([self.modifications_indicator, self.modifications_ignore]),
        ]

        self.version_selector = VersionSelectorWidget()
        ipw.dlink((self.app, 'available_release_lines'), (self.version_selector.release_line, 'options'),
                  transform=lambda release_lines: [(self._format_release_line_name(release_line), release_line)
                                                   for release_line in release_lines])
        ipw.dlink((self.app, 'installed_release_line'), (self.version_selector.release_line, 'value'),
                  transform=lambda value: value if value else None)
        ipw.dlink((self.app, 'installed_version'), (self.version_selector.installed_version, 'value'),
                  transform=lambda version: '' if version is None else version)
        self.version_selector.release_line.observe(self._refresh_widget_state, names=['value'])
        ipw.dlink((self.app, 'busy'), (self.version_selector, 'disabled'))
        self.version_selector.layout.visibility = 'visible' if with_version_selector else 'hidden'
        self.version_selector.disabled = True
        children.insert(1, self.version_selector)

        super().__init__(children=children)

        self.app.observe(self._refresh_widget_state)
        self.app.refresh_async()  # init all widgets

    @staticmethod
    def _format_release_line_name(release_line):
        """Return a human-readable version of a release line name."""
        if re.match(r'git:refs\/heads\/.*', release_line):
            return re.sub(r'git:refs\/heads\/', '', release_line)
        return release_line

    def _refresh_widget_state(self, _=None):
        """Refresh the widget to reflect the current state of the app."""
        with self.hold_trait_notifications():
            # Collect information about app state.
            busy = self.app.busy
            modified = self.app.modified
            override = modified and self.modifications_ignore.value
            blocked = modified and not self.modifications_ignore.value

            # Prepare warning icons and messages depending on whether we override or not.
            # These messages and icons are only shown if needed.
            warn_or_ban_icon = "ban" if blocked else "warning"
            if override:
                tooltip = "Operation will lead to potential loss of local modifications!"
            else:
                tooltip = "Operation blocked due to local modifications."

            # Determine whether we can install, updated, and uninstall.
            requested_release_line = self.version_selector.release_line.value
            switch_release_line = requested_release_line is not None \
                and requested_release_line != self.app.installed_release_line
            can_install = switch_release_line or not self.app.is_installed()
            can_uninstall = self.app.is_installed()
            try:
                can_update = self.app.updates_available and not can_install
            except RuntimeError:
                can_update = None

            # Update the install button state.
            self.install_button.disabled = busy or blocked or not can_install
            self.install_button.button_style = 'info' if can_install else ''
            self.install_button.icon = '' if can_install and not modified else warn_or_ban_icon if can_install else ''
            self.install_button.tooltip = '' if can_install and not modified else tooltip if can_install else ''
            if switch_release_line:
                self.install_button.description = f'Install ({self._format_release_line_name(requested_release_line)})'
            else:
                self.install_button.description = 'Install'

            # Update the uninstall button state.
            self.uninstall_button.disabled = busy or blocked or not can_uninstall
            self.uninstall_button.button_style = 'danger' if can_uninstall else ''
            self.uninstall_button.icon = \
                "" if can_uninstall and not modified else warn_or_ban_icon if can_uninstall else ""
            self.uninstall_button.tooltip = '' if can_uninstall and not modified else tooltip if can_uninstall else ''

            # Update the update button state.
            self.update_button.disabled = busy or blocked or not can_update
            if self.app.is_installed() and can_update is None:
                self.update_button.icon = 'warning'
                self.update_button.tooltip = 'Unable to determine availability of updates.'
            else:
                self.update_button.icon = \
                    "circle-up" if can_update and not modified else warn_or_ban_icon if can_update else ""
                self.update_button.button_style = 'success' if can_update else ''
                self.update_button.tooltip = '' if can_update and not modified else tooltip if can_update else ''

            # Indicate whether there are local modifications and present option for user override.
            self.modifications_indicator.value = \
                f'<i class="fa fa-{warn_or_ban_icon}"> There are local modifications.' if modified else ''
            self.modifications_ignore.layout.visibility = 'visible' if modified else 'hidden'

    def _show_msg_success(self, msg):
        """Show a message indicating successful execution of a requested operation."""
        self.install_info.show_temporary_message(HTML_MSG_SUCCESS.format(msg))

    def _show_msg_failure(self, msg):
        """Show a message indicating failure to execute a requested operation."""
        self.install_info.show_temporary_message(HTML_MSG_FAILURE.format(msg))

    def _check_modified_state(self):
        """Check whether there are local modifications to the app that prevent any install etc. operations."""
        self.app.refresh()
        self._refresh_widget_state()
        blocked = self.app.modified and not self.modifications_ignore.value
        if blocked:
            raise RuntimeError("Unable to perform operation, there are local modifications to the app repository.")

    def _install_version(self, _):
        """Attempt to install the a specific version of the app."""
        release_line = self.version_selector.release_line.value
        try:
            self._check_modified_state()
            release_line = self.app.install_app(release_line)  # argument may be None
        except (AssertionError, RuntimeError, CalledProcessError) as error:
            self._show_msg_failure(str(error))
        else:
            self._show_msg_success(f"Installed app ({self._format_release_line_name(release_line)}).")

    def _update_app(self, _):
        """Attempt to uninstall the app."""
        try:
            self._check_modified_state()
            self.app.update_app()
        except (AssertionError, RuntimeError, CalledProcessError) as error:
            self._show_msg_failure(str(error))
        else:
            self._show_msg_success("Updated app.")

    def _uninstall_app(self, _):
        """Attempt to uninstall the app."""
        try:
            self._check_modified_state()
            self.app.uninstall_app()
        except RuntimeError as error:
            self._show_msg_failure(str(error))
        else:
            self._show_msg_success("Uninstalled app.")



# -*- coding: utf-8 -*-
"""AiiDA lab app store."""

import ipywidgets as ipw
from IPython.display import display
from jinja2 import Template

from .app import AiidaLabApp, AppManagerWidget
from .config import AIIDALAB_APPS
from .utils import load_app_registry


class AiidaLabAppStore(ipw.HBox):
    """Class to manage AiiDA lab app store."""

    def __init__(self):
        requested_dict = load_app_registry()
        if requested_dict:
            self.registry_sorted_list = sorted(requested_dict['apps'].items())
            categories_dict = requested_dict['categories']
        else:
            self.registry_sorted_list = []
        self.app_corresponding_categories = []
        self.output = ipw.Output()

        # Apps per page.
        self.items_per_page = ipw.BoundedIntText(
            value=10,
            min=5,
            max=40,
            step=5,
            description='Apps per page:',
            disabled=False,
            style={'description_width': 'initial'},
            layout={'width': '150px'},
        )
        self.items_per_page.observe(self.update_page_selector, names='value')

        # Page selector.
        self.page_selector = ipw.ToggleButtons(
            options=[],
            description='Page:',
            disabled=False,
            style={'button_width': '30px'},
        )
        self.page_selector.observe(self.render, names='value')

        # Only installed filter.
        self.only_installed = ipw.Checkbox(value=False, description='Show only installed', disabled=False)
        self.only_installed.observe(self.change_vis_list, names='value')

        # Category filter.
        self.category_filter = ipw.SelectMultiple(
            options=[],
            value=[],
            #rows=10,
            description='Filter categories',
            style={'description_width': 'initial'},
            disabled=False)
        apply_category_filter = ipw.Button(description="apply filter")
        apply_category_filter.on_click(self.change_vis_list)
        clear_category_filter = ipw.Button(description='clear selection')
        clear_category_filter.on_click(self._clear_category_filter)
        self.category_title_key_mapping = {
            categories_dict[key]['title'] if key in categories_dict else key: key for key in categories_dict
        }
        self.category_filter.options = list(self.category_title_key_mapping)

        # Define the apps that are going to be displayed.
        self.apps_to_display = [AiidaLabApp(name, app, AIIDALAB_APPS) for name, app in self.registry_sorted_list]

        self.update_page_selector()
        super().__init__([
            self.only_installed,
            ipw.VBox([self.category_filter,
                      ipw.HBox([apply_category_filter, clear_category_filter])])
        ])

    def _clear_category_filter(self, _):
        self.category_filter.value = ()
        self.change_vis_list()

    def update_page_selector(self, _=None):
        """This function changes the current page value and makes sure that the page is re-rendered."""
        initial_page = self.page_selector.value
        self.page_selector.options = list(range(1, int(len(self.apps_to_display) / self.items_per_page.value + 2)))

        # this parts makes sure that render function will always run one time only:
        # if page number changed in the previous step - the render was run automatically already
        # if not -it needs to be done manually
        if initial_page == self.page_selector.value:
            self.render()

    def change_vis_list(self, _=None):
        """This function creates a list of apps to be displayed. Moreover, it creates a parallel list of categories.
        After this the page selector update is called."""
        self.apps_to_display = [AiidaLabApp(name, app, AIIDALAB_APPS) for name, app in self.registry_sorted_list]

        if self.only_installed.value:
            self.apps_to_display = [app for app in self.apps_to_display if app.is_installed()]

        if self.category_filter.value:
            all_apps = self.apps_to_display
            self.apps_to_display = []  # clear the array that contains all the apps to be displayed
            self.app_corresponding_categories = []  # create a parallel array that contains corresponding category names
            # iterate over all categories
            for category in self.category_filter.value:
                category_key = self.category_title_key_mapping[category]
                apps_belonging_to_category = [app for app in all_apps if app.in_category(category_key)]
                self.apps_to_display += apps_belonging_to_category
                if apps_belonging_to_category:
                    self.app_corresponding_categories += [category] + [None] * (len(apps_belonging_to_category) - 1)

        # As the list of apps to be shown is changed, it is essential to update the page selector also
        self.update_page_selector()

    def render(self, _=None):
        """Show all the available apps."""
        self.output.clear_output()
        page = self.page_selector.value - 1
        start = page * self.items_per_page.value
        end = (page + 1) * self.items_per_page.value
        with self.output:
            for number, app_base in enumerate(self.apps_to_display[start:end]):
                if self.category_filter.value and self.app_corresponding_categories[start:end][number]:
                    display(ipw.HTML("<h1>{}</h1>".format(
                        self.app_corresponding_categories[start:end][number].title())))

                widget = AppStoreAppManagerWidget(app_base)
                display(ipw.HTML('<hr>'))  # horizontal line
                display(widget)

        return self.output


class AppStoreAppManagerWidget(AppManagerWidget):

    TEMPLATE = Template("""<h2 style="text-align:center;">{{ app.title }}</h2>
    <div style="font-size: 15px; font-style: italic">Description: {{ app.description | truncate(200) }}</div>
    <br>
    <div style="text-align:right;"><a href=./single_app.ipynb?app={{ app.name }}>Manage App</a>""")



# -*- coding: utf-8 -*-
"""AiiDA lab basic widgets."""

from threading import Timer

import traitlets
import ipywidgets as ipw


class _StatusWidgetMixin:
    """Show temporary messages for example for status updates.

    This is a mixin class that is meant to be part of an inheritance
    tree of an actual widget with a 'value' traitlet that is used
    to convey a status message. See the non-private classes below
    for examples.
    """

    def __init__(self, *args, **kwargs):
        self._clear_timer = None
        super().__init__(*args, **kwargs)

    def _clear_value(self):
        """Set widget .value to be an empty string."""
        self.value = ''

    def show_temporary_message(self, value, clear_after=3):
        """Show a temporary message and clear it after the given interval."""
        if self._clear_timer is not None:
            # Cancel previous timer; has no effect if it already timed out.
            self._clear_timer.cancel()

        self.value = value

        # Start new timer that will clear the value after the specified interval.
        self._clear_timer = Timer(clear_after, self._clear_value)
        self._clear_timer.start()


class StatusLabel(_StatusWidgetMixin, ipw.Label):
    """Show temporary messages for example for status updates."""


class StatusHTML(_StatusWidgetMixin, ipw.HTML):
    """Show temporary HTML messages for example for status updates."""


class UpdateAvailableInfoWidget(ipw.HTML):
    """Widget that indicates whether an update is available."""

    updates_available = traitlets.Bool(allow_none=True)

    MESSAGES = {
        None:
            """<font color="#D8000C"><i class='fa fa-times-circle'></i> """\
            """Unable to determine availability of updates.</font>""",
        True:
            """<font color="#9F6000"><i class='fa fa-warning'></i> Update Available</font>""",
        False:
            """<font color="#270"><i class='fa fa-check'></i> Latest Version</font>""",
    }

    def __init__(self, updates_available=None, **kwargs):
        super().__init__(updates_available=None, **kwargs)
        self._observe_updates_available(dict(new=updates_available))  # initialize

    @traitlets.observe('updates_available')
    def _observe_updates_available(self, change):
        self.value = self.MESSAGES[change['new']]


class Spinner(ipw.HTML):
    """Widget that shows a simple spinner if enabled."""

    enabled = traitlets.Bool()

    def __init__(self, spinner_style=None):
        self.spinner_style = f' style="{spinner_style}"' if spinner_style else ''
        super().__init__()

    @traitlets.default('enabled')
    def _default_enabled(self):  # pylint: disable=no-self-use
        return False

    @traitlets.observe('enabled')
    def _observe_enabled(self, change):
        """Show spinner if enabled, otherwise nothing."""
        if change['new']:
            self.value = f"""<i class="fa fa-spinner fa-pulse"{self.spinner_style}></i>"""
        else:
            self.value = ""
